/*
    $Id: MappedProductReader.java,v 1.4 2007-07-16 07:21:55 ralf Exp $

    Copyright (c) 2006 Brockmann Consult. All rights reserved. Use is
    subject to license terms.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the Lesser GNU General Public License as
    published by the Free Software Foundation; either version 2 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with the BEAM software; if not, download BEAM from
    http://www.brockmann-consult.de/beam/ and install it.
*/
package org.esa.beam.globcolour.dataio;

import org.esa.beam.dataio.netcdf.NetcdfReader;
import org.esa.beam.dataio.netcdf.NetcdfReaderPlugIn;
import org.esa.beam.framework.dataio.AbstractProductReader;
import org.esa.beam.framework.dataio.ProductReader;
import org.esa.beam.framework.datamodel.Band;
import org.esa.beam.framework.datamodel.Product;
import org.esa.beam.framework.datamodel.ProductData;

import java.io.IOException;

/**
 * The class <code>MappedProductReader</code> provides a reader for the Mapped products
 * generated by the GlobColour processor.
 * <p/>
 * Instances of this class are created only by the <code>createReaderInstance()</code>
 * method of a {@link MappedProductReaderPlugIn} object.
 * <p/>
 * This implementation overrides the <code>readProductNodes()</code> and <code>readBandData()</code>
 * methods of the {@link AbstractProductReader}.
 *
 * @author Ralf Quast
 * @version $Revision: 1.4 $ $Date: 2007-07-16 07:21:55 $
 */
public class MappedProductReader extends AbstractProductReader {

    private ProductReader productReader;

    /**
     * Constructs an instance of this class.
     *
     * @param productReaderPlugIn the plug-in which creates this reader instance.
     */
    public MappedProductReader(MappedProductReaderPlugIn productReaderPlugIn) {
        super(productReaderPlugIn);

        productReader = new NetcdfReader(new NetcdfReaderPlugIn());
    }

    /**
     * Factory method for creating a {@link Product} by reading the metadata from
     * the underlying netCDF file.
     * <p/>
     * This method is called only once during the lifetime of this reader instance.
     * It is invoked as the last step in the <code>readProductNodes()</code> method
     * of the {@link AbstractProductReader} base class.
     *
     * @throws java.io.IOException if an I/O error occurs.
     */
    @Override
    protected Product readProductNodesImpl() throws IOException {
        final Product product = productReader.readProductNodes(getInput(), getSubsetDef());

        if (ProductUtilities.isDiagnosticDataSet(product)) {
            product.setProductType(ProductTypes.MAPPED_DDS);
        } else {
            product.setProductType(ProductTypes.MAPPED_GLOBAL);
        }

        product.setProductReader(this);
        ProductUtilities.extend(product);
        product.setModified(false);

        return product;
    }

    /**
     * Implementation of the template method which is called after an optional spatial subset
     * has been defined.
     * <p/>
     * The target band, buffer and region parameters are exactly those that have been passed
     * to the original call. The target offset parameters have been taken into acount by the
     * source offset parameters, so an implementor of this method is free to ignore these.
     *
     * @param sourceOffsetX the X-offset for the source raster subset.
     * @param sourceOffsetY the Y-offset for the source raster subset.
     * @param sourceWidth   the width of the source raster subset.
     * @param sourceHeight  the height of source raster subset.
     * @param sourceStepX   the sub-sampling in X direction within the source raster subset.
     * @param sourceStepY   the sub-sampling in Y direction within the region providing samples to be read.
     * @param targetBand    the target band.
     * @param targetOffsetX the X-offset for the target raster (can be ignored).
     * @param targetOffsetY the Y-offset for the target raster (can be ignored).
     * @param targetWidth   the width of target raster.
     * @param targetHeight  the height of the target raster.
     * @param targetBuffer  the buffer which stores the sample values read.
     *
     * @throws IOException if an I/O error occurs
     * @see #readBandRasterData
     * @see #getSubsetDef
     */
    @Override
    protected void readBandRasterDataImpl(int sourceOffsetX,
                                          int sourceOffsetY,
                                          int sourceWidth,
                                          int sourceHeight,
                                          int sourceStepX,
                                          int sourceStepY,
                                          Band targetBand,
                                          int targetOffsetX,
                                          int targetOffsetY,
                                          int targetWidth,
                                          int targetHeight,
                                          ProductData targetBuffer)
            throws IOException {
        productReader.readBandRasterData(targetBand, targetOffsetX, targetOffsetY, targetWidth, targetHeight,
                                         targetBuffer);
    }

    @Override
    public void close() throws IOException {
        if (productReader != null) {
            productReader.close();
        }
        super.close();
    }

}
