package org.esa.beam.globcolour.internal;

import static java.lang.Math.sqrt;
import static java.lang.Math.atan2;
import static java.lang.Math.toRadians;
import static java.lang.Math.cos;
import static java.lang.Math.sin;
import static java.lang.Math.asin;
import java.util.Calendar;
import java.util.GregorianCalendar;
import java.util.TimeZone;

/**
 * Created by IntelliJ IDEA.
 *
 * @author Ralf Quast
 * @version $Revision: 1.2 $ $Date: 2007-07-16 07:21:55 $
 */
public class MatchupUtilities {

    /**
     * Based on formulae developed by Flandern & Pulkkinen (1979, Astrophys. J. Suppl. Ser. 41, 391)
     * and simplified by P. Schlyter. The zenith angle is calculated with an accuracy of 1 arcmin.
     *
     * @param calendar the time of observation
     * @param lat      the geographical latitude of the observer (degrees)
     * @param lon      the geographical longitude of the observer (degrees)
     *
     * @return the zenith and azimuth angles of the Sun (degrees, azimuth angle not implemented yet)
     */
    static SunPosition computeSunPosition(final Calendar calendar, final double lat, final double lon) {
        final Calendar utc = toUTC(calendar);

        final int year = utc.get(Calendar.YEAR);
        final int month = utc.get(Calendar.MONTH) + 1;
        final int date = utc.get(Calendar.DATE);
        final int hour = utc.get(Calendar.HOUR_OF_DAY);
        final int minute = utc.get(Calendar.MINUTE);
        final double h = hour + minute / 60.0;

        // fractional day, 0.0 corresponds to January 1, 2000 00:00:00 UTC
        final double d = h / 24.0 + (367 * year - 7 * (year + (month + 9) / 12) / 4 + 275 * month / 9 + date - 730530);
        // ecliptical obliquity
        final double o = toRadians(23.4393) - toRadians(0.0000003563) * d;
        // mean anomaly
        final double M = toRadians(356.0470) + toRadians(0.9856002585) * d;
        // longitude of perihelion
        final double w = toRadians(282.9404) + toRadians(0.0000470935) * d;
        // eccentricity
        final double e = 0.016709 - 1.151E-9 * d;
        // eccentric anomaly
        final double E = M + e * sin(M) * (1.0 + e * cos(M));
        // true anomaly
        final double v = atan2(sqrt(1.0 - e * e) * sin(E), cos(E) - e);
        // true longitude
        final double l = v + w;

        // Cartesian geocentric coordinates
        final double x = cos(l);
        final double y = sin(l) * cos(o);
        final double z = sin(l) * sin(o);

        // right ascension
        final double alpha = atan2(y, x);
        // declination
        final double delta = atan2(z, sqrt(x * x + y * y));
        // local siderial time
        final double lst = M + w + Math.PI + h * toRadians(15.0) + toRadians(lon);
        // local hour angle
        final double lha = lst - alpha;
        // latitude (rad)
        final double phi = toRadians(lat);
        // altitude angle
        final double alt = asin(sin(phi) * sin(delta) + cos(phi) * cos(delta) * cos(lha));
        // azimuth angle
        final double xaa = sin(lha) * cos(delta);
        final double yaa = cos(phi) * sin(delta) - cos(lha) * cos(delta) * sin(phi) + cos(phi) * sin(delta);
        final double saa = atan2(xaa, yaa);

        // zenith angle
        final double sza = Math.PI / 2.0 - alt;

        return new SunPosition(Math.toDegrees(sza), Math.toDegrees(saa));
    }

    static Calendar toUTC(Calendar calendar) {
        final Calendar utc = new GregorianCalendar(TimeZone.getTimeZone("UTC"));
        utc.setTimeInMillis(calendar.getTimeInMillis());

        return utc;
    }
}
